/*
 *	Wireless Tools
 *
 *		Jean II - HPLB 97->99 - HPL 99->04
 *
 * Main code for "iwconfig". This is the generic tool for most
 * manipulations...
 * You need to link this code against "iwlib.c" and "-lm".
 *
 * This file is released under the GPL license.
 *     Copyright (c) 1997-2004 Jean Tourrilhes <jt@hpl.hp.com>
 */

#include "iwlib.h"		/* Header */

//#define SIOCDEVPRIVATE                              0x8BE0 
#define RTPRIV_IOCTL_GET_MAC_TABLE					(SIOCIWFIRSTPRIV + 0x0F)
#define RTPRIV_IOCTL_WSC_PROFILE                    (SIOCIWFIRSTPRIV + 0x12)
#define RT_PRIV_IOCTL				(SIOCIWFIRSTPRIV + 0x01)

#define OID_GET_SET_TOGGLE				0x8000

#define RT_OID_WSC_QUERY_STATUS						0x0751

#define IOCTL_GET_MAC_TABLE	0
#define IOCTL_GET_WSC_PROFILE  1
#define IOCTL_GET_WSC_STATUS   2

#define MAX_NUMBER_OF_MAC 50
#define MAC_ADDR_LENGTH 6

#define ULONG unsigned long
#define USHORT unsigned short
#define UCHAR unsigned char
#define UINT32 unsigned int

/******************************* MAIN ********************************/
// MIMO Tx parameter, ShortGI, MCS, STBC, etc.  these are fields in TXWI. Don't change this definition!!!
typedef union   _MACHTTRANSMIT_SETTING {
        struct       {
        USHORT          MCS:7;                 // MCS
        USHORT          BW:1;      //channel bandwidth 20MHz or 40 MHz
        USHORT          ShortGI:1;
        USHORT          STBC:2;    //SPACE 
        USHORT          rsv:3;       
        USHORT          MODE:2; // Use definition MODE_xxx.  
        }       field;
        USHORT          word;
 } MACHTTRANSMIT_SETTING, *PMACHTTRANSMIT_SETTING;

typedef struct _RT_802_11_MAC_ENTRY {
    UCHAR       Addr[MAC_ADDR_LENGTH];
    UCHAR       Aid;
    UCHAR       Psm;     // 0:PWR_ACTIVE, 1:PWR_SAVE
    UCHAR                MimoPs;  // 0:MMPS_STATIC, 1:MMPS_DYNAMIC, 3:MMPS_Enabled
    char           AvgRssi0;
        char               AvgRssi1;
        char               AvgRssi2;
        UINT32            ConnectedTime;
    MACHTTRANSMIT_SETTING       TxRate;
    MACHTTRANSMIT_SETTING       MaxTxRate;
        UCHAR            apidx;
} RT_802_11_MAC_ENTRY, *PRT_802_11_MAC_ENTRY;

typedef struct _RT_802_11_MAC_TABLE {
    ULONG       Num;
    RT_802_11_MAC_ENTRY Entry[MAX_NUMBER_OF_MAC];
} RT_802_11_MAC_TABLE, *PRT_802_11_MAC_TABLE;

#define PACKED  __attribute__ ((packed))
#define USHORT  unsigned short
#define UCHAR   unsigned char
typedef struct PACKED _WSC_CONFIGURED_VALUE {
    USHORT WscConfigured; // 1 un-configured; 2 configured
    UCHAR   WscSsid[32 + 1];
    USHORT WscAuthMode; // mandatory, 0x01: open, 0x02: wpa-psk, 0x04: shared, 0x08:wpa, 0x10: wpa2, 0x
    USHORT  WscEncrypType;  // 0x01: none, 0x02: wep, 0x04: tkip, 0x08: aes
    UCHAR   DefaultKeyIdx;
    UCHAR   WscWPAKey[64 + 1];
} WSC_CONFIGURED_VALUE;

static int
iw_sockets_open(void)
{
  static const int families[] = {
    AF_INET, AF_IPX, AF_AX25, AF_APPLETALK
  };
  unsigned int	i;
  int		sock;

  /*
   * Now pick any (exisiting) useful socket family for generic queries
   * Note : don't open all the socket, only returns when one matches,
   * all protocols might not be valid.
   * Workaround by Jim Kaba <jkaba@sarnoff.com>
   * Note : in 99% of the case, we will just open the inet_sock.
   * The remaining 1% case are not fully correct...
   */

  /* Try all families we support */
  for(i = 0; i < sizeof(families)/sizeof(int); ++i)
    {
      /* Try to open the socket, if success returns it */
      sock = socket(families[i], SOCK_DGRAM, 0);
      if(sock >= 0)
	return sock;
  }

  return -1;
}

static void swap(void *p, int size)
{
	int i,j;
	unsigned char *tmp = (unsigned char *)p;
	unsigned char char_buf;
	
	for(i = size, j = 0; i > 0; i = i-2,j++)
	{
		char_buf = tmp[j];
		tmp[j] = tmp[i+j-1];
		tmp[i+j-1] = char_buf;
	}
}
/*
WscAuthMode     // 0x01: open, 0x02: wpa-psk, 0x04: shared, 0x08:wpa, 0x10: wpa2, 0x20: wpa2psk;0x2200: wpa2Mixed
WscEncrypType   // 0x01: none, 0x02: wep, 0x04: tkip, 0x08: aes ,0x0C: tkipaes
*/
static void parse_packet(char *data, int cmd, int todo)
{
	//unsigned int i;

	if(todo == IOCTL_GET_MAC_TABLE)
	{
		RT_802_11_MAC_TABLE *mp;
		int i;
		mp = (RT_802_11_MAC_TABLE *)data;
		printf("\n%-4s%-20s%-4s%\n",
				"AID", "MAC_Address", "PSM");
		for(i = 0 ; i < mp->Num ; i++)
		{
			printf("%-4d", mp->Entry[i].Aid);
			printf("[%02X:%02X:%02X:%02X:%02X:%02X] ",
			mp->Entry[i].Addr[0], mp->Entry[i].Addr[1],
			mp->Entry[i].Addr[2], mp->Entry[i].Addr[3],
			mp->Entry[i].Addr[4], mp->Entry[i].Addr[5]);
			printf("%-4d", mp->Entry[i].Psm);
			printf("\n");
		}
	}

	if(todo == IOCTL_GET_WSC_STATUS)
	{
		int status;

		memcpy(&status,data,sizeof(status));
		printf("%d\n",status);
	}
	if(todo == IOCTL_GET_WSC_PROFILE)
	{
		
		WSC_CONFIGURED_VALUE *profile;
		profile = (WSC_CONFIGURED_VALUE *)data;
		
		/*{
		//for debug
		char *pppp;
		int iiii = 0;
		pppp = (char *)data;
		for( iiii = 0; iiii < 500; iiii++)
		{
			if(iiii % 16 == 0)
			{
				printf("\n");	
			}				
			printf("%02x",*(unsigned char*)(pppp + iiii));

		}
	
		}
		printf("\n");*/
		if(profile->WscConfigured == 0x0100)
			printf("un-configured\2");
		else
			printf("configured\2");
		printf("%s\2", profile->WscSsid);
//		printf("WscAuthMode=");
		switch(profile->WscAuthMode)
		{
			case 0x0100: printf("open\2");
					   break;
		    case 0x0200: printf("psk\2");
		    			break;
		    case 0x0400: printf("shared\2");
		    			break;
		    case 0x0800: printf("wpa\2");
		    			break;
		    case 0x1000: printf("wpa2\2");
		    			break;
		    case 0x2000: printf("psk2\2");
		    			break;
		    case 0x2200: printf("psk2_mixed\2");
		    			break;
		    default:
		    	printf("NONE\2");
		}
//		printf("WscEncrypType=");
		switch(profile->WscEncrypType)
		{
			case 0x0100: printf("none\2");
					   break;
		    case 0x0200: printf("wep\2");
		    			break;
		    case 0x0400: printf("tkip\2");
		    			break;
		    case 0x0800: printf("aes\2");
		    			break;
		    case 0x0c00: printf("tkip+aes\2");
		    			break;
		    default:
		    	printf("NONe\2");
		}
		printf("%d\2", profile->DefaultKeyIdx);
		printf("%s\2", profile->WscWPAKey);
	}
}

/*------------------------------------------------------------------*/
/*
 * The main !
 */
int
main(int	argc,
     char **	argv)
{
  int skfd;		/* generic raw socket desc.	*/
  int goterr = 0;
  unsigned int cmd = 0;
  unsigned short flags = 0;
  char data[2048];
  char name[100];
  int todo = 0;
  struct iwreq wrq;

  if(argc != 3)
  	return -1;

  /* Create a channel to the NET kernel. */
  if((skfd = iw_sockets_open()) < 0)
    {
      perror("socket");
      return(-1);
    }

  if(strcmp(argv[2],"get_mac_table") == 0)
  {
  	cmd = RTPRIV_IOCTL_GET_MAC_TABLE;
	todo = IOCTL_GET_MAC_TABLE;
  }
  if(strcmp(argv[2],"get_wsc_profile") == 0)
  {
 	cmd = RTPRIV_IOCTL_WSC_PROFILE;
 	flags = 0;
	todo = IOCTL_GET_WSC_PROFILE;
  }
  if(strcmp(argv[2],"get_wsc_status") == 0)
  {
  	cmd = RT_PRIV_IOCTL;
	flags = RT_OID_WSC_QUERY_STATUS;
	todo = IOCTL_GET_WSC_STATUS;
  }

  if(cmd == 0)
  	return -1;

  sprintf(name, argv[1]);                  
  strcpy(data, argv[2]);
  strcpy(wrq.ifr_name, name);
  wrq.u.data.length = strlen(data);
  wrq.u.data.pointer = data;
  wrq.u.data.flags = flags;
  goterr = ioctl(skfd, cmd, &wrq);
  if(goterr == 0)
  	parse_packet(data,cmd,todo);
		
  exit(0);
}


